/* NOTE: copyright notice at end of file.
 */

#include <flash_helper.h>
#include <nrf_mesh_config_app.h>

#include <bbn_debugging.h>
#include <device_state_manager.h>
#include <flash_manager.h>
#include <log.h>
#include <mesh_app_utils.h>
#include <mesh_stack.h>


/**** Flash handling ****/
#if PERSISTENT_STORAGE

static flash_manager_t m_flash_manager;

volatile unsigned int app_flash_pending;



typedef void (*afh_exec_t)(void);
/**
 * @brief execute a deferred function passed as an argument
 */
static void afh_exec_callback(void *p_args)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Flash: callback exec\n");
    afh_exec_t fptr = (afh_exec_t)p_args;
    fptr();
}



static void afh_reset(void)
{
    static fm_mem_listener_t mem_listener = {
        .callback = afh_exec_callback,
        .p_args = afh_reset
    };

    /* Register the listener and wait for some memory to be freed up before we retry. */
    if (flash_manager_remove(&m_flash_manager) != NRF_SUCCESS)
        flash_manager_mem_listener_register(&mem_listener);

    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Flash: removing data\n");
}



static void afh_remove_complete(const flash_manager_t *p_manager)
{
    app_flash_pending--;
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Flash: remove complete (%u)\n", app_flash_pending);
}

static void afh_invalidate_complete(const flash_manager_t *p_manager, fm_handle_t handle, fm_result_t result)
{
    /* This application does not expect invalidate complete calls. */
    BBN_ERROR_IF(true, "Flash: unexpected invalidate\n");
}

static void afh_write_complete(const flash_manager_t *p_manager, const fm_entry_t *p_entry, fm_result_t result)
{
    /* If we get an AREA_FULL then our calculations for flash space required are buggy. */
    BBN_ERROR_CHECK(result == FM_RESULT_ERROR_AREA_FULL);

    /* We do not invalidate in this module, so a NOT_FOUND should not be received. */
    BBN_ERROR_CHECK(result == FM_RESULT_ERROR_NOT_FOUND);

    /* Malfunction should also trigger a reset. */
    BBN_ERROR_CHECK(result == FM_RESULT_ERROR_FLASH_MALFUNCTION);

    app_flash_pending--;
     __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Flash: write complete (%u)\n", app_flash_pending);
}



void app_flash_init(void)
{
    static fm_mem_listener_t flash_add_mem_available_struct = {
        .callback = afh_exec_callback,
        .p_args = app_flash_init
    };

    app_flash_pending = 1;

    const uint32_t *start_address;
    uint32_t allocated_area_size;
    ERROR_CHECK(
            mesh_stack_persistence_flash_usage(&start_address, &allocated_area_size)
            );

    flash_manager_config_t manager_config = {
        .write_complete_cb = afh_write_complete,
        .invalidate_complete_cb = afh_invalidate_complete,
        .remove_complete_cb = afh_remove_complete,
        .min_available_space = WORD_SIZE,
        .p_area = (const flash_manager_page_t *)((uint32_t)start_address - PAGE_SIZE * APP_FLASH_PAGE_COUNT),
        .page_count = APP_FLASH_PAGE_COUNT
    };

    uint32_t status = flash_manager_add(&m_flash_manager, &manager_config);
    if (NRF_SUCCESS != status)
    {
        flash_manager_mem_listener_register(&flash_add_mem_available_struct);
        return;
    }

    app_flash_pending = 0;
    __LOG(LOG_SRC_APP, LOG_LEVEL_ERROR, "Flash: init complete\n");
}

uint32_t app_flash_data_load(fm_handle_t entry_handle, void *p_data, uint16_t length)
{
    uint32_t read_len = length;

    uint32_t status = flash_manager_entry_read(&m_flash_manager, entry_handle, p_data, &read_len);
    if (status != NRF_SUCCESS)
        memset(p_data, 0x00, length);

    return NRF_SUCCESS;
}

uint32_t app_flash_data_store(fm_handle_t entry_handle, const void *p_data, uint16_t length)
{
    /* Sanity: maximum data length */
    if (length > FLASH_MANAGER_ENTRY_MAX_SIZE)
    {
        __LOG(LOG_SRC_APP, LOG_LEVEL_ERROR,
                "Flash write length %d > FLASH_MANAGER_ENTRY_MAX_SIZE %d\n",
                length, FLASH_MANAGER_ENTRY_MAX_SIZE);
        return NRF_ERROR_INVALID_PARAM;
    }

    /* Sanity: data must be 32-bit aligned else flash manager will fail _silently_
     *
     * If you see this warning you are writing a data length which is not
     * a multiple of 4 (32-bit aligned).
     * If your data is a struct, you can use '__attribute__((aligned(4)))';
     * beware this will pad (increase) the size of your struct.
     */
    if (length & 0x3)
    {
        __LOG(LOG_SRC_APP, LOG_LEVEL_ERROR, "Flash write length %d is not 32-bit aligned\n", length);
        return NRF_ERROR_INVALID_PARAM;
    }

    fm_entry_t *p_entry = flash_manager_entry_alloc(&m_flash_manager, entry_handle, length);
    if (!p_entry)
        return NRF_ERROR_BUSY;

    app_flash_pending++;
    memcpy(p_entry->data, p_data, length);
    flash_manager_entry_commit(p_entry);
    return NRF_SUCCESS;
}

void app_flash_clearall()
{
    app_flash_pending++;
    afh_reset();
}

void app_flash_clear(void *p_data, uint16_t length)
{
    memset(p_data, 0x00, length);
    app_flash_clearall(); /* existing applications rely on this old (incorrect) behavior */
}




#else

void app_flash_init(void)
{ ; }

uint32_t app_flash_data_load(fm_handle_t entry_handle, void *p_data, uint16_t length)
{
    memset(p_data, 0x00, length);
    return NRF_ERROR_NOT_SUPPORTED;
}

uint32_t app_flash_data_store(fm_handle_t entry_handle, const void *p_data, uint16_t length)
{
    return NRF_ERROR_NOT_SUPPORTED;
}

void app_flash_clear(void *p_data, uint16_t length)
{
    memset(p_data, 0x00, length);
}

void app_flash_clearall()
{ ; }

#endif



/* Copyright (c) 2010 - 2019, Nordic Semiconductor ASA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
