#ifndef FLASH_HELPER_H__
#define FLASH_HELPER_H__

/**
 * NOTE: copyright notice at bottom of file.
 */

#include <bbn_utilities.h>
#include <flash_manager.h>
#ifndef APP_FLASH_PAGE_COUNT
    #define APP_FLASH_PAGE_COUNT 1
#endif


/**
 * @brief gives number of pending flash operations _without_
 * calling into the mesh stack with flash_manager_is_stable() (there be dragons)
 */
extern volatile unsigned int app_flash_pending;


/**
 * Loads the application specific data from flash
 *
 * Data is loaded to `p_data` if entry is found, else `p_data` is filled with `0x00`.
 *
 * @param[in]  entry_handle Entry handle
 * @param[out] p_data       Pointer to the application data structure which will hold the data
 * @param[in]  length       Length of the data structure
 *
 * @return NRF_SUCCESS          If entry was found and data was loaded
 * @return NRF_ERROR_NOT_FOUND  If entry is not found
 */
uint32_t app_flash_data_load(fm_handle_t entry_handle, void *p_data, uint16_t length);

/**
 * Stores given data structure to the flash
 *
 * @note Data structure cannot be longer than @ref FLASH_MANAGER_ENTRY_MAX_SIZE
 * @note Data structure must be 32-bit aligned (see definition for details)
 *
 * @param[in]  entry_handle Entry handle
 * @param[in]  p_data        Pointer to the application data structure which will be stored
 * @param[in]  length       Length of the data structure
 *
 * @return NRF_ERROR_BUSY   If flash is busy and no more store requests can be queued
 * @return NRF_SUCCESS      If entry is committed succesfully
 */
uint32_t app_flash_data_store(fm_handle_t entry_handle, const void *p_data, uint16_t length);

MACRO_INLINE void app_flash_stabilize()
{
    /* sometimes _not_ stable even though nothing pending */
    while (app_flash_pending || !flash_manager_is_stable())
    {
#if NRF_SDH_DISPATCH_MODEL == NRF_SDH_DISPATCH_MODEL_INTERRUPT
        ;
#elif NRF_SDH_DISPATCH_MODEL == NRF_SDH_DISPATCH_MODEL_APPSH
        nrf_mesh_process();
        app_sched_execute();
#endif
    }
}

/**
 * Adds the flash manger instance for the application.
 *
 * Call this function during bootup.
 */
void app_flash_init(void);

/**
 * Set 'length' bytes at *p_data to 0; clear _all_ flash managed by flash_manager.
 *
 * Clears _all_ flash memory being managed (not just that being referenced in the call),
 * hence the compile-time warning - you probably do not want to call this!
 */
__attribute__((warning("Removes _all_ data from flash. You probably do not want this!")))
void app_flash_clear(void *p_data, uint16_t length);

/**
 * Removes _all_ entries from flash
 */
void app_flash_clearall();


/* Copyright (c) 2010 - 2019, Nordic Semiconductor ASA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#endif /* FLASH_HELPER_H__ */
